
class PyleaseError(Exception):
    def __init__(self, message='', *args, **kwargs):
        super(PyleaseError, self).__init__(*args, **kwargs)

        self.message = message


class VersionSpecError(PyleaseError):
    pass


class ReleaseError(PyleaseError):
    pass


class UploadError(PyleaseError):
    pass



from __future__ import print_function
from abc import ABCMeta, abstractmethod
import os
from pylease.cmd.rollback import Rollback, Stage

from pylease.logger import LOGME as logme  # noqa
from pylease.ctxmgmt import Caution
from pylease.ex import PyleaseError
from pylease.filemgmt import update_files, VersionRollback
from pylease.releasemgmt import release
from pylease.util import SubclassIgnoreMark


class Command(object):
    # pylint: disable=no-self-use, too-many-instance-attributes
    # The whole logic of Pylease is centralized on the Command class,
    # thus it is reasonable to have more than seven instance attributes.
    """
    This class is one of the main point of Pylease.
    For adding new commands just inherit from this class and implement :func:`~pylease.cmd.Command._process_command` method.
    """

    __metaclass__ = ABCMeta

    _IGNORE_ME_VAR_NAME = 'ignore_me'
    ignore_me = False

    def __init__(self, lizy, name, description, rollback=None, requires_project=True):
        """
        This constructor should be called from child classes and at least be supplied with at least ``name`` and ``description``.

        Arguments:
            lizy (pylease.Pylease): The `lizy` object, which is initialized and passed by Pylease.
            name (str): The name of the command, which will appear in the `usage` output.
            description (str): Description of the command which will also appear in the help message.
            rollback (pylease.cmd.rollback.Rollback): The rollback object that will be executed in case of failure during or after the
                command. This parameter may be emitted if the command does not need a rollback, or may be set in the process of command
                execution using the :func:`~pylease.cmd.task.BeforeTask.enable_rollback` method, if it depends on some parameters during
                runtime.
            requires_project (bool): Boolean indicating whether the command requires to operate on an existing project. E.g. the ``init``
                command requires an empty directory.
        """
        super(Command, self).__init__()

        logme.debug("Initializing %s command with rollback %s", name, rollback)

        self.name = name
        self.description = description
        self.parser = lizy.add_subparser(self.name, help=self.description)
        self._lizy = lizy
        self.before_tasks = set()
        self.after_tasks = set()
        self.result = None
        self.rollback = rollback
        self.requires_project = requires_project

        lizy.add_command(self.name, self)

    @abstractmethod
    def _process_command(self, lizy, args):
        """
        The method which should be implemented when inheriting the :class:`~pylease.cmd.Command`. All the command logic must go into this
        method.

        Arguments:
            lizy (pylease.Pylease): The :class:`~pylease.Pylease` singleton.
            args (argparse.Namespace): The arguments passed to the command line while invoking Pylease.
        """
        pass  # pragma: no cover

    def __call__(self, args):
        with Caution(verbose=args.verbose) as caution:
            if self.requires_project and self._lizy.info_container.is_empty:
                raise PyleaseError("'{}' command requires an existing project!".format(self.name))  # pragma: no cover

            logme.debug("Executing before tasks.")
            for task in self.before_tasks:
                rollback = task(self._lizy, args)
                logme.debug("Before task returned rollback %s.", rollback)
                caution.add_rollback(rollback)

            logme.debug("Executing command %s with args %s and rollback %s.", self.name, args, self.rollback)
            result = self._process_command(self._lizy, args)
            logme.debug("Command %s finished, rollback is %s.", self.name, self.rollback)
            caution.add_rollback(self.rollback)

            self.result = result

            logme.debug("Executing after tasks.")
            for task in self.after_tasks:
                rollback = task(self._lizy, args)
                logme.debug("After task returned rollback %s.", rollback)
                caution.add_rollback(rollback)

        return caution.result

    @classmethod
    def init_all(cls, lizy):
        commands = cls.__subclasses__()

        for command in commands:
            command.init_all(lizy)
            if not getattr(command, cls._IGNORE_ME_VAR_NAME):
                command(lizy)

    def add_before_task(self, task):
        """
        Adds a :class:`~pylease.cmd.task.BeforeTask` to the :class:`~pylease.cmd.Command`.

        Arguments:
            task (pylease.cmd.task.BeforeTask): The task to be added.
        """
        self.before_tasks.add(task)
        task.set_command(self)

    def add_after_task(self, task):
        """
        Adds a :class:`~pylease.cmd.task.AfterTask` to the :class:`~pylease.cmd.Command`.

        Arguments:
            task (pylease.cmd.task.AfterTask): The task to be added.
        """
        self.after_tasks.add(task)
        task.set_command(self)


class NamedCommand(Command):
    # pylint: disable=W0223, W1401, too-few-public-methods
    # NamedCommand is also abstract
    # The number of public methods is reasonable for this kind of class
    """
    Same as the :class:`~pylease.cmd.Command` class, however this class enables a little taste of convenience. You can define a class
    having name with a suffix "\ |pylease_named_command_suffix|\ " and it will automatically assign the prefix of the class name as the
    command name.
    """

    _SUFFIX = "Command"
    ignore_me = SubclassIgnoreMark('NamedCommand')

    def __init__(self, lizy, description, rollback=None, requires_project=True):
        """
        Same as :func:`~pylease.cmd.Command.__init__` of :class:`~pylease.cmd.Command` class, except that the ``name`` argument is passed
        automatically.
        """
        my_name = self.__class__.__name__
        name = my_name[:-(len(self._SUFFIX))].lower()

        super(NamedCommand, self).__init__(lizy, name, description, rollback, requires_project)


class StatusCommand(NamedCommand):
    # pylint: disable=too-few-public-methods
    # The number of public methods is reasonable for this kind of class

    OUTPUT_FMT = 'Project Name: {name}\nCurrent Version: {version}'

    KEY_VERSION = 'version'
    KEY_NAME = 'name'

    def __init__(self, lizy):
        super(StatusCommand, self).__init__(lizy, 'Retrieve current status of the project')

    def _process_command(self, lizy, args):
        print(self.OUTPUT_FMT.format(name=lizy.info_container.name, version=lizy.info_container.version))

        return {self.KEY_NAME: lizy.info_container.name, self.KEY_VERSION: lizy.info_container.version}


class MakeCommand(NamedCommand):
    # pylint: disable=too-few-public-methods
    # The number of public methods is reasonable for this kind of class

    KEY_OLD_VERSION = 'old_version'
    KEY_NEW_VERSION = 'new_version'
    KEY_LEVEL = 'level'

    def __init__(self, lizy):
        super(MakeCommand, self).__init__(lizy, 'Make a release')

        release_group = self.parser.add_argument_group(title='release arguments',
                                                       description='Specify one of those arguments to make the '
                                                       'corresponding level release')

        level_group = release_group.add_mutually_exclusive_group(required=True)
        level_group.add_argument('--major', dest='level', action='store_const', const='major', help='Make a major release')
        level_group.add_argument('--minor', dest='level', action='store_const', const='minor', help='Make a minor release')
        level_group.add_argument('--patch', dest='level', action='store_const', const='patch', help='Make a patch release')
        level_group.add_argument('--dev', dest='level', action='store_const', const='dev', help='Make a dev release')

    def _process_command(self, lizy, args):
        old_version = lizy.info_container.version

        if old_version is None:
            error_msg = "Version specification not found!"
            logme.error(error_msg)
            raise PyleaseError(error_msg)

        level = args.level

        new_version = release(old_version, level)
        logme.info('%s releasing %s to v%s', level.capitalize(), lizy.info_container.name, new_version)

        files = lizy.get_version_files()

        counts = update_files(old_version, new_version, files)

        count = 0
        for version_file in counts:
            count += counts[version_file]

        if count > 1:
            logme.warn("More than one version specification found.")
            logme.debug("Occurrences:")
            for filename in counts:
                msg = "\t{filename}: {count}".format(filename=filename, count=counts[filename])
                logme.debug(msg)

        self.rollback = VersionRollback(old_version, new_version, files).rollback
        return {self.KEY_OLD_VERSION: str(old_version), self.KEY_NEW_VERSION: str(new_version), self.KEY_LEVEL: str(level)}


class InitCommand(NamedCommand):
    # pylint: disable=too-few-public-methods
    # The number of public methods is reasonable for this kind of class

    SETUP_PY_CONTENTS = """import {name}
from setuptools import setup

setup(name='{name}',
      version={name}.__version__)
"""
    SETUP_CFG_CONTENTS = """[pylease]
version-files = {name}/__init__.py
"""
    INIT_PY_CONTENTS = """__version__ = '{version}'
"""
    INITIAL_VERSION = '0.0'

    SETUP_PY_NAME = 'setup.py'
    SETUP_CFG_NAME = 'setup.cfg'
    INIT_PY_NAME_FMT = '{name}/__init__.py'

    def __init__(self, lizy):
        super(InitCommand, self).__init__(lizy, 'Initialize a new Python project', requires_project=False)

        self.parser.add_argument('name', action='store', type=str, help='name of the project to be initiated')

    def _process_command(self, lizy, args):
        super(InitCommand, self)._process_command(lizy, args)

        project_name = args.name
        setup_py_name = self.SETUP_PY_NAME
        setup_cfg_name = self.SETUP_CFG_NAME
        init_py_name = self.INIT_PY_NAME_FMT.format(name=project_name)

        setup_py_contents = self.SETUP_PY_CONTENTS.format(name=args.name)
        setup_cfg_contents = self.SETUP_CFG_CONTENTS.format(name=args.name)
        init_py_contents = self.INIT_PY_CONTENTS.format(version=self.INITIAL_VERSION)

        rollback = InitRollback(project_name)
        self.rollback = rollback

        nodes_in_cwd = [node for node in os.listdir('.') if not node.startswith('.')]
        if nodes_in_cwd:  # pragma: no cover
            logme.debug('Directory not empty.\nFollowing files are present:\n%s', nodes_in_cwd)
            raise PyleaseError('Directory not empty.\nSorry, but the directory must be empty to initialise a new project.')

        os.mkdir(project_name)
        rollback.enable_stage(InitRollback.DIRECTORY_STAGE)

        self._write_file(setup_py_name, setup_py_contents)
        rollback.enable_stage(rollback.SETUPPY_STAGE)

        self._write_file(setup_cfg_name, setup_cfg_contents)
        rollback.enable_stage(rollback.SETUPCFG_STAGE)

        self._write_file(init_py_name, init_py_contents)
        rollback.enable_stage(rollback.INITPY_STAGE)

    @classmethod
    def _write_file(cls, filename, contents):
        with open(os.path.join(os.getcwd(), filename), 'w') as the_file:
            the_file.write(contents)


class InitRollback(Rollback):  # pragma: no cover; Tested in the scope of Rollback class
    SETUPPY_STAGE = 'setuppy'
    SETUPCFG_STAGE = 'setupcfg'
    INITPY_STAGE = 'initpy'
    DIRECTORY_STAGE = 'directory'

    def __init__(self, project_name):
        super(InitRollback, self).__init__()

        self._name = project_name

    @Stage(SETUPPY_STAGE)
    def rollback_setuppy(self):
        logme.debug('Removing setup.py')
        os.remove(InitCommand.SETUP_PY_NAME)

    @Stage(SETUPCFG_STAGE)
    def rollback_setupcfg(self):
        logme.debug('Removing setup.cfg')
        os.remove(InitCommand.SETUP_CFG_NAME)

    @Stage(INITPY_STAGE)
    def rollback_initpy(self):
        logme.debug('Removing __init__.py')
        os.remove(InitCommand.INIT_PY_NAME_FMT.format(name=self._name))

    @Stage(DIRECTORY_STAGE)
    def rollback_directory(self):
        logme.debug('Removing %s', self._name)
        os.remove(self._name)



from unittest import TestCase
from nose.tools import eq_, ok_
from pylease import releasemgmt
from pylease.ex import ReleaseError

__author__ = 'bagrat'


class TestReleaseManagement(TestCase):
    def test_release_must_update_the_version_object_depending_on_the_level_provided(self):
        raises_error = False
        try:
            releasemgmt.release('1.0', 'unknown')
        except ReleaseError:
            raises_error = True

        ok_(raises_error)

        expected_version = '1.2.3.dev4'
        eq_(str(releasemgmt.release('1.2.3.dev3', 'dev')), expected_version)

        expected_version = '1.2.4'
        eq_(str(releasemgmt.release('1.2.3.dev3', 'patch')), expected_version)

        expected_version = '1.3'
        eq_(str(releasemgmt.release('1.2.3.dev3', 'minor')), expected_version)

        expected_version = '2.0'
        eq_(str(releasemgmt.release('1.2.3.dev3', 'major')), expected_version)



from cStringIO import StringIO
import sys
import textwrap
from unittest import TestCase
import shutil
import __builtin__
import mock
import os

__author__ = 'bagrat'


class TestError(Exception):
    pass


class PyleaseTest(TestCase):
    MOCK_SUBDIR = os.path.join('tests', 'mock')

    def setUp(self):  # noqa
        super(PyleaseTest, self).setUpClass()

        cwd = os.getcwd()
        self.mock_path = os.path.join(cwd, self.MOCK_SUBDIR)

        if os.path.exists(self.mock_path):
            shutil.rmtree(self.MOCK_SUBDIR)

        os.makedirs(self.mock_path)

    def tearDown(self):  # noqa
        super(PyleaseTest, self).tearDownClass()

        shutil.rmtree(self.MOCK_SUBDIR)


class MockedFile(object):
    def __init__(self, filename, content, test=None, mock_path=None, for_import=False):
        super(MockedFile, self).__init__()

        if not mock_path:
            if not test:
                raise TestError("While mocking file without a mock path, you have to provide the test class.")
            if not issubclass(test.__class__, PyleaseTest):
                raise TestError("While mocking a file you have provided '{}' as test class which is not a subclass of PyleaseTest, "
                                "though it is supposed to be.". format(test.__name__))

            mock_path_attr_name = 'mock_path'
            if not hasattr(test, mock_path_attr_name):
                raise TestError("While mocking file you have provided '{}' as test class which does not have '{}' attribute, "
                                "though it is supposed to have.". format(test.__name__, mock_path_attr_name))

            mock_path = getattr(test, mock_path_attr_name)

        self._mock_path = mock_path
        self._content = content
        self._filename = filename
        self.mock_file_path = os.path.join(self._mock_path, self._filename)
        self._for_import = for_import
        self._orig_open = None
        self._orig_getcwd = None

    def __enter__(self):
        with open(self.mock_file_path, 'w') as file_mock:
            file_mock.write(self._content)

        self._orig_open = getattr(__builtin__, 'open')
        __builtin__.open = self._open_mock()

        if self._for_import:
            self._orig_getcwd = os.getcwd
            os.getcwd = mock.Mock(return_value=self._mock_path)

            sys.path = [os.getcwd()] + sys.path

            if 'setup' in sys.modules:
                del sys.modules['setup']

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        __builtin__.open = self._orig_open

        if self._for_import:
            os.getcwd = self._orig_getcwd

    def contents(self):
        with self._orig_open(self.mock_file_path, 'r') as mock_file:
            return mock_file.read()

    def _open_mock(self):
        mocked_abs_file = os.path.abspath(self._filename)

        def open_mock(filename, *args, **kwargs):
            if self._for_import:
                os.getcwd = self._orig_getcwd
            provided_abs_file = os.path.abspath(filename)
            if self._for_import:
                os.getcwd = mock.Mock(return_value=self._mock_path)

            if mocked_abs_file == provided_abs_file:
                return self._orig_open(self.mock_file_path, *args, **kwargs)
            return self._orig_open(filename, *args, **kwargs)

        return open_mock


class MockedSetupPy(MockedFile):
    FILENAME = 'setup.py'

    def __init__(self, content, test=None, mock_path=None, for_import=True):
        if for_import:
            content = textwrap.dedent(content)

        super(MockedSetupPy, self).__init__(self.FILENAME, content, test, mock_path, for_import)


class CapturedStdout(object):
    def __init__(self):
        super(CapturedStdout, self).__init__()

        self._stdout = None
        self.output = None
        self._stringio = None

    def __enter__(self):
        self._stdout = sys.stdout
        sys.stdout = self._stringio = StringIO()

        return self

    def __exit__(self, *args):
        self.output = self._stringio.getvalue()

        sys.stdout = self._stdout


class MockedFileWrite(object):
    _file_map = {}

    def __init__(self, filename):
        super(MockedFileWrite, self).__init__()

        self._filename = filename
        self._io = StringIO()
        self._orig_open = __builtin__.open
        self._file_map[filename] = self._io

    def open_mock(self):
        def mock(filename, *args, **kwargs):
            if filename in self._file_map:
                return self._file_map[filename]
            else:
                return self._orig_open(filename, *args, **kwargs)

        return mock

    def __enter__(self):
        __builtin__.open = self.open_mock()

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        __builtin__.open = self._orig_open

        del self._file_map[self._filename]

    def contents(self):
        return self._io.getvalue()



from subprocess import call
import subprocess
from pylease.cmd.rollback import Rollback, Stage
from pylease.ex import PyleaseError
from pylease.cmd.task import AfterTask
from pylease.ext import Extension
from pylease.logger import LOGME as logme  # noqa


class GitExtension(Extension):
    """
    Git extension creates an annotated tag in the git repository named
    v<version>, where version is the current version being released. Also
    makes additional commit containing the updated version files.
    """
    # pylint: disable=too-few-public-methods
    # The number of public methods is reasonable for this kind of class
    def load(self):
        make_command = self._lizy.commands['make']
        parser = make_command.parser

        git_group = parser.add_argument_group(title='git arguments')

        git_group.add_argument('--git-tag',
                               action='store_true',
                               dest='use_git',
                               default=False,
                               help='Create a version tag on git')

        make_command.add_after_task(GitAfterTask())


class GitAfterTask(AfterTask):  # pragma: no cover - Unable to test this other than manually TODO: try to
    # pylint: disable=too-few-public-methods
    TAG_MESSAGE_FMT = 'Prepare release v{version}'

    def execute(self, lizy, args):
        if args.use_git:
            version = self._command_result[self._command.KEY_NEW_VERSION]
            tag_name = 'v{version}'.format(version=version)

            logme.info('Creating git tag %s', tag_name)

            self.enable_rollback(GitRollback(version))

            logme.debug("Staging %s files for commit", lizy.get_version_files())
            proc = subprocess.Popen(['git', 'add'] + lizy.get_version_files(), stderr=subprocess.PIPE)
            proc.wait()
            if proc.returncode:
                err = proc.stderr.read()
                raise PyleaseError(err)

            logme.debug("Committing files %s", lizy.get_version_files())
            proc = subprocess.Popen(['git', 'commit', '-m', self.TAG_MESSAGE_FMT.format(version=version)], stderr=subprocess.PIPE)
            proc.wait()
            if proc.returncode:
                err = proc.stderr.read()
                raise PyleaseError(err)
            self._rollback.enable_stage(GitRollback.COMMIT_STAGE)

            logme.debug("Adding git tag")
            proc = subprocess.Popen(['git', 'tag', '-a', tag_name], stderr=subprocess.PIPE)
            proc.wait()
            if proc.returncode:
                err = proc.stderr.read()
                raise PyleaseError(err)
            self._rollback.enable_stage(GitRollback.TAG_STAGE)


class GitRollback(Rollback):  # pragma: no cover - Unable to test this other than manually TODO: try to test
    COMMIT_STAGE = 'commit'
    TAG_STAGE = 'tag'

    def __init__(self, version):
        super(GitRollback, self).__init__()

        self._version = version

    @Stage(COMMIT_STAGE, 0)
    def rollback_commit(self):
        tag_message = GitAfterTask.TAG_MESSAGE_FMT.format(version=self._version)
        proc = subprocess.Popen('git log -n 1 --grep="{}"  --format="%H"'.format(tag_message), stdout=subprocess.PIPE, shell=True)
        proc.wait()
        release_commit = proc.stdout.read()

        proc = subprocess.Popen('git log -n 2 --format="%H" {}'.format(release_commit), stdout=subprocess.PIPE, shell=True)
        proc.wait()
        pre_release_commit = proc.stdout.read().split('\n')[1]

        proc = subprocess.Popen(['git', 'reset', pre_release_commit])
        proc.wait()

    @Stage(TAG_STAGE, 1)
    def rollback_tag(self):
        call(['git', 'tag', '-d', 'v{version}'.format(version=self._version)])



from mock import Mock, MagicMock
from nose.tools import ok_

from pylease.ctxmgmt import Caution, ReplacedSetup
from tests import PyleaseTest, MockedSetupPy


class ContextManagersTest(PyleaseTest):
    def test_replaced_setup_must_replace_the_setuptools_setup_with_provided_callback(self):
        key1 = 'key1'
        val1 = 'val1'
        key2 = 'key2'
        val2 = 'val2'

        kwargs = {key1: val1, key2: val2}
        setup_py = """
                   from setuptools import setup

                   kwargs = {{'{}': '{key1}', '{}': '{key2}'}}
                   setup(**kwargs)
                   """. format(key1, key2, **kwargs)

        callback = Mock()

        with ReplacedSetup(callback):
            with MockedSetupPy(setup_py, self):
                __import__('setup')

            callback.assert_called_once_with(**kwargs)

    def test_caution_context_manager_must_rollback_everything_if_error_occurs(self):
        rb1 = MagicMock()
        rb2 = MagicMock()
        rb3 = MagicMock()

        with Caution() as caution:
            caution.add_rollback(rb1)
            caution.add_rollback(rb2)

            raise Exception()

        rb1.assert_called_once_with()
        rb2.assert_called_once_with()
        ok_(not rb3.called)

    def test_caution_context_manager_should_leave_everythin_as_is_if_no_error_occurs(self):
        rb1 = MagicMock()

        with Caution() as caution:
            caution.add_rollback(rb1)

        ok_(not rb1.called)



from unittest import TestCase
from nose.tools import eq_, ok_
from pylease.vermgmt import DevedVersion
from pylease import InfoContainer

__author__ = 'bagrat'


class TestVersionManagement(TestCase):
    def test_deved_version_must_increase_corresponding_version_level_depending_on_method_called(self):
        ver = DevedVersion('1.1')

        eq_(ver.version, (1, 1, 0))

        ver = DevedVersion('1.2.dev45')

        eq_(ver.version, (1, 2, 0))
        eq_(ver.dev, 45)

        ver = DevedVersion('2.3.1.dev78')

        eq_(str(ver), '2.3.1.dev78')

        ver = DevedVersion('1.2.3.dev4')

        ver.increase_major()
        ver.increase_minor()
        ver.increase_patch()
        ver.increase_dev()

        eq_(str(ver), '2.1.1.dev1')

        raises_value_error = False
        try:
            DevedVersion('wrongversion')
        except ValueError:
            raises_value_error = True

        ok_(raises_value_error)

    def test_info_container_set_info_method_must_set_the_objects_attributes_to_the_values_from_the_dictionary(self):
        ic = InfoContainer()

        info = {'key1': 'val1', 'key2': 'val2'}

        ic.set_info(**info)

        for key in info:
            eq_(getattr(ic, key), info[key])



from pylease.ex import ReleaseError
from pylease.vermgmt import DevedVersion

__author__ = 'bagrat'

_LEVELS = ['major', 'minor', 'patch', 'dev']
_DEFAULT_LEVEL_INDEX = 0
DEFAULT_LEVEL = _LEVELS[_DEFAULT_LEVEL_INDEX]


def release(old, level=DEFAULT_LEVEL, count=1):
    """
    Returns a new version which is ``count`` times higher from ``old`` by level ``level``.

    :param old: The current version
    :param level: The level to be released to, i.e. one of the following - major, minor, patch, dev
    :param count: The number to increase the level by
    """

    if level not in _LEVELS:
        raise ReleaseError('Unknown release level: {level}\nShould be one of those: {levels}'.format(level=level, levels=_LEVELS))

    new_version = DevedVersion(old)

    if level == 'major':
        new_version.increase_major(count)
    elif level == 'minor':
        new_version.increase_minor(count)
    elif level == 'patch':
        new_version.increase_patch(count)
    elif level == 'dev':
        new_version.increase_dev(count)

    return new_version



import os
import sys
from argparse import ArgumentParser


import pylease
import pylease.logger as logger
from pylease.ext import Extension
from pylease.ext import git, pypi  # noqa
from pylease.cmd import Command
from pylease.ctxmgmt import ReplacedSetup
from pylease import InfoContainer

__author__ = 'bagrat'


def main(args=None):
    """
    The main entry point of Pylease utility. Called when Pylease is launched from command line. May also be called from another python
    package by just supplying command line arguments as a list.

    :param args: List of command line arguments
    :return: Error code indicating completion status
    """
    # Initialize argument parser
    parser = ArgumentParser()
    parser.add_argument('--version', action='version', version=version_info())
    parser.add_argument('--verbose', '-v',
                        action='store_true',
                        dest='verbose',
                        default=False,
                        help='Make Pylease verbose, i.e. show more execution information')
    sub_parsers = parser.add_subparsers(help='Pylease commands', dest='command')

    # Collect current information about the project
    info = InfoContainer()
    if os.path.exists(os.path.join(os.getcwd(), 'setup.py')):
        sys.path = [os.getcwd()] + sys.path
        with ReplacedSetup(info.set_info):
            __import__('setup')
    else:
        info.is_empty = True  # pragma: no cover

    # Initialize Lizy
    lizy = pylease.Pylease(parser, sub_parsers, info)

    # Initialize Commands and Extensions
    Command.init_all(lizy)
    Extension.init_all(lizy)

    # Parse and handle arguments
    args = parser.parse_args(args)

    if args.verbose:
        logger.add_verbose_handler()

    command_name = args.command

    # Execute requested command and return error code
    return lizy.execute_command(command_name, args)


def version_info():
    """
    Generates output of Pylease version information, when it is called with --version argument

    :return: Version information string
    """
    version_str = "Pylease version {version}"
    version = pylease.__version__

    return version_str.format(version=version)



import logging


class ColorFormatter(logging.Formatter):  # pragma: no cover
    # pylint: disable=too-few-public-methods
    """
    Custom Formatter for Pylease logger that colors the log message depending on the log level.
    """
    color_end = "\x1b[1;0m"
    colors = dict(
        default='0',
        grey='30',
        green="32",
        red="31",
        yellow="33",
        purple="35",
        cyan="36"
    )
    level_map = {
        logging.DEBUG: ('cyan', False),
        logging.INFO: ('green', False),
        logging.WARNING: ('yellow', False),
        logging.ERROR: ('red', False),
        logging.CRITICAL: ('red', True)
    }

    def __init__(self, fmt=None, datefmt=None):
        super(ColorFormatter, self).__init__(fmt, datefmt)

    def format(self, record):
        lvl = int(record.levelno / 10) * 10
        record.msg = self._colorize(self.level_map.get(lvl)[0], record.msg, self.level_map.get(lvl)[1])
        return super(ColorFormatter, self).format(record)

    def _colorize(self, color, text, bold=False):
        """
        Borrowed from nosetests spec plugin by bitprophet https://github.com/bitprophet/spec/blob/master/spec/plugin.py
        """
        bold = 1 if bold else 0
        return "\x1b[%s;%sm%s%s" % (bold, self.colors.get(color, self.colors['default']), text, self.color_end)


class LevelFilter(logging.Filter):  # pragma: no cover
    # pylint: disable=too-few-public-methods
    def __init__(self, name='', level=logging.DEBUG):
        super(LevelFilter, self).__init__(name)

        self._level = level

    def filter(self, record):
        return int(record.levelno / 10) * 10 == self._level


_LOGGING_FMT = "%(message)s"

LOGME = logging.getLogger(__name__)
LOGME.setLevel(logging.DEBUG)

HANDLER = logging.StreamHandler()
HANDLER.setLevel(logging.INFO)
HANDLER.setFormatter(ColorFormatter(fmt=_LOGGING_FMT))

LOGME.addHandler(HANDLER)


def add_verbose_handler():  # pragma: no cover
    """
    Adds separate handler to the Pylease logger with level DEBUG.
    """
    verbose_handler = logging.StreamHandler()
    verbose_handler.setLevel(logging.DEBUG)
    verbose_handler.setFormatter(ColorFormatter(fmt=_LOGGING_FMT))

    verbose_handler.addFilter(LevelFilter(level=verbose_handler.level))

    LOGME.addHandler(verbose_handler)



import ConfigParser
from pylease.logger import LOGME as logme  # noqa
import os

import pylease.cmd
import pylease.ext

__version__ = '0.3'


class Pylease(object):
    """
    The main class of Pylease, which contains all the needed resources for extensions. This class is initialised once and by Pylease,
    which is the so called ``lizy`` object. It is passed to :class:`~pylease.cmd.Command` and :class:`~pylease.ext.Extension` instances.

    Attributes:
        info_container (pylease.InfoContainer): Contains information about current status of the project. Minimal information is
            ``name`` and ``version``.
        commands (dict): A dictionary of Pylease commands, including commands defined in extensions if any. The values of the dictionary
            are instances of :class:`~pylease.cmd.Command` class.
        parser (argparse.ArgumentParser): The root parser of Pylease. Use this object to add command line arguments to Pylease on the
            same level as ``--version`` and ``--help``.
        config (dict): A dictionary representing the configuration parsed from ``setup.cfg`` defined under ``[pylease]`` section. If a
            configuration value in the configuration file is defined as ``key1 = valA, valB, valC`` then the value of the ``key1`` key of
            this attribute will be an instance of :class:`~list` and be equal to ``['valA', 'valB, 'valC']``.
    """
    def __init__(self, parser, cmd_subparsers, info_container):
        super(Pylease, self).__init__()

        self.parser = parser
        self.cmd_subparsers = cmd_subparsers
        self.info_container = info_container
        self.commands = {}

        config = {}
        config_parser = ConfigParser.SafeConfigParser()
        try:
            if os.path.exists('setup.cfg'):
                config_parser.read('setup.cfg')
                items = config_parser.items('pylease')
                for item in items:
                    config[item[0]] = item[1]
        except ConfigParser.NoSectionError:
            logme.warn('No pylease section found in setup.cfg')

        self.config = config

        self._load_extensions()

    def add_command(self, name, command):
        self.commands[name] = command

    def execute_command(self, name, args):
        return self.commands[name](args)

    def add_subparser(self, *args, **kwargs):
        return self.cmd_subparsers.add_parser(*args, **kwargs)

    def _get_config_list_value(self, key):
        values = None
        if key in self.config:
            values_str_list = self.config[key].replace(' ', '')
            values = values_str_list.split(',')

        return values

    def get_version_files(self):
        return self._get_config_list_value('version-files')

    def get_plugins(self):
        return self._get_config_list_value('use-plugins') or []

    def _load_extensions(self):
        extension_packages = self.get_plugins()

        for package in extension_packages:
            __import__(package)


class InfoContainer(object):
    # pylint: disable=too-few-public-methods
    """
    A simple container that maps a provided dictionary to its attributes. This provides the current status of the project,
    and the minimal built-in information attributes are the following:

    Attributes:
        name (str): The name of the project.
        version (str): The current versin of the project
        is_empty (bool): The status of current working directory, i.e. indicates whether it is empty or not.
    """
    def __init__(self):
        super(InfoContainer, self).__init__()

        self.name = None
        self.version = None
        self.is_empty = False

    def set_info(self, **kwargs):
        """
        Used to extend the information about the project.

        Example:
            Below are two options on how to use/extend the ``InfoContainer``::

                info = InfoContainer()

                # Option 1
                info.set_info(info1='value2', info2='value2')

                # Option 2
                more_info = {'info3': 'value3'}
                info.set_info(**more_info)

                # Then you can access your info as instance attributes
                print(info.info2)  # will print 'value2'

        """
        for key in kwargs:
            setattr(self, key, kwargs[key])



from abc import ABCMeta, abstractmethod
from pylease.ctxmgmt import Caution
from pylease.logger import LOGME as logme  # noqa


class BeforeTask(object):
    """

    """
    # pylint: disable=too-few-public-methods
    # The number of public methods is reasonable for this kind of class

    __metaclass__ = ABCMeta

    def __init__(self, rollback=None):
        super(BeforeTask, self).__init__()

        logme.debug("Initializing %s with rollback '%s'", self.__class__.__name__, rollback)
        self._command = None
        self._rollback = rollback
        self._needs_rollback = False

    def enable_rollback(self, rollback=None):
        """
        Enables a rollback for the task. The provided rollback gets executed in case of failure in the execution stack.

        Attributes:
            rollback (pylease.cmd.rollback.Rollback): The rollback instance for the task.
        """
        if rollback:
            if self._rollback:
                logme.debug('Overwriting existing rollback for task "%s"', self.__class__.__name__)
            self._rollback = rollback
        self._needs_rollback = True

    def set_command(self, command):
        self._command = command

    def __call__(self, lizy, args):
        try:
            logme.debug("Executing task %s", self.__class__.__name__)
            self.execute(lizy, args)

            if self._needs_rollback:
                logme.debug("Done executing task %s, rollback is %s", self.__class__.__name__, self._rollback)
                return self._rollback
        except BaseException as ex:
            if self._needs_rollback:
                logme.debug("%s error occurred, setting rollback '%s' to exception", ex.__class__.__name__, self._rollback)
                setattr(ex, Caution.EXCEPTION_ROLLBACK_ATTR_NAME, self._rollback)
            raise ex

    @abstractmethod
    def execute(self, lizy, args):
        """
        The place where the extension logic goes on.

        Arguments:
            lizy (pylease.Pylease): The :class:`~pylease.Pylease` singleton that provides all the needed information about the project.
            args (argparse.Namespace): The arguments supplied to the command line.
        """
        pass  # pragma: no cover


class AfterTask(BeforeTask):
    # pylint: disable=too-few-public-methods, W0223
    # AfterTask is also abstract
    # The number of public methods is reasonable for this kind of class
    __metaclass__ = ABCMeta

    @property
    def _command_result(self):
        """
        A dictionary containing information by the completion of the command execution.
        """
        return self._command.result  # pragma: no cover



from traceback import format_tb
import setuptools
import sys
from pylease.logger import LOGME as logme  # noqa


class Caution(object):
    # pylint: disable=too-few-public-methods
    # The number of public methods is reasonable for this kind of class
    """
    Context manager for handling rollback process in case of pylease failure
    """

    EXCEPTION_ROLLBACK_ATTR_NAME = 'rollback'

    def __init__(self, verbose=False):
        super(Caution, self).__init__()

        self._rollbacks = set()
        self.result = 0
        self._verbose = verbose

    def add_rollback(self, rollback):
        if rollback:
            self._rollbacks.add(rollback)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type:
            logme.error("Some error occurred: rolling back...\n%s", exc_val.message)
            logme.debug("Error occurred with type %s", exc_type.__name__)
            if self._verbose:
                logme.debug('\n'.join(format_tb(exc_tb)))

            logme.debug("Checking %s exception for rollback", exc_type.__name__)
            if hasattr(exc_val, self.EXCEPTION_ROLLBACK_ATTR_NAME):
                rollback = getattr(exc_val, self.EXCEPTION_ROLLBACK_ATTR_NAME)
                if rollback:
                    logme.debug("Found rollback '%s', executing...", rollback)
                    rollback()
                else:
                    logme.debug("No rollback provided")  # pragma: no cover - no logic involved here, just logging

            for rollback in self._rollbacks:
                rollback()

            self.result = 1

        return True


class ReplacedSetup(object):
    """
    Context manager for replacing setuptools setup method and then setting
    all back.
    """
    # pylint: disable=too-few-public-methods
    # The number of public methods is reasonable for this kind of class
    def __init__(self, callback):
        super(ReplacedSetup, self).__init__()

        self._callback = callback
        self._old_setup = None

    def __enter__(self):
        self._old_setup = setuptools.setup
        setuptools.setup = self._version_reporter

    def __exit__(self, exc_type, exc_val, exc_tb):
        setuptools.setup = self._old_setup
        if 'setup' in sys.modules:
            del sys.modules['setup']

    def _version_reporter(self, **kwargs):
        """
        The replacement method for setup method.
        """
        self._callback(**kwargs)



import textwrap
from mock import MagicMock, mock_open, patch
from nose.tools import eq_, ok_
import sys
import pylease
from pylease.cmd import StatusCommand, InitCommand
from pylease.ext import Extension
from pylease.main import main
from tests import PyleaseTest, MockedSetupPy, CapturedStdout, MockedFile, MockedFileWrite


class CommandLineTest(PyleaseTest):
    def test_status_command_must_show_project_info(self):
        name = 'Some Project'
        version = 'Some Version'

        setup_py = """
                   from setuptools import setup

                   setup(name='{name}', version='{version}')
                   """

        with MockedSetupPy(setup_py.format(name=name, version=version), self):
            with CapturedStdout() as stdout:
                main(['status'])

        eq_(StatusCommand.OUTPUT_FMT.format(name=name, version=version) + '\n', stdout.output)

    def test_make_release_must_update_the_version_correspondingly(self):
        version = '0.12'
        setup_py_contents = textwrap.dedent("""
                                            from setuptools import setup

                                            setup(version='{version}')
                                            """)

        with MockedSetupPy(setup_py_contents.format(version=version), self) as setup_py:
            with MockedFile('setup.cfg', '', self):
                main(['make', '--dev'])
                expected_version = '0.12.dev1'
                eq_(setup_py.contents(), setup_py_contents.format(version=expected_version))

        with MockedSetupPy(setup_py_contents.format(version=version), self) as setup_py:
            with MockedFile('setup.cfg', '', self):
                main(['make', '--patch'])
                expected_version = '0.12.1'
                eq_(setup_py.contents(), setup_py_contents.format(version=expected_version))

        with MockedSetupPy(setup_py_contents.format(version=version), self) as setup_py:
            with MockedFile('setup.cfg', '', self):
                main(['make', '--minor'])
                expected_version = '0.13'
                eq_(setup_py.contents(), setup_py_contents.format(version=expected_version))

        with MockedSetupPy(setup_py_contents.format(version=version), self) as setup_py:
            with MockedFile('setup.cfg', '', self):
                main(['make', '--major'])
                expected_version = '1.0'
                eq_(setup_py.contents(), setup_py_contents.format(version=expected_version))

    def test_make_command_must_return_error_when_no_version_spec_is_found(self):
        setup_py_contents = textwrap.dedent("""
                                            from setuptools import setup

                                            setup()
                                            """)
        with MockedSetupPy(setup_py_contents, self):
            ok_(main(['make', '--major']) != 0)

    def test_make_command_must_warn_wwhen_more_than_one_version_specs_are_found(self):
        pylease.logger.LOGME.warn = MagicMock()

        setup_py_contents = textwrap.dedent("""
                                            from setuptools import setup
                                            version='1.0'
                                            setup(version='1.0')
                                            """)

        with MockedSetupPy(setup_py_contents, self):
            with MockedFile('setup.cfg', '', self):
                main(['make', '--major'])

        ok_(pylease.logger.LOGME.warn.called)

    def test_pylease_must_load_configuration_into_pylease_object(self):
        key1 = 'key1'
        key2 = 'key2'
        val1 = 'val1'
        val2 = 'val2'
        contents = textwrap.dedent("""
                [pylease]
                {} = {}
                {} = {}
                   """.format(key1, val1, key2, val2))
        with MockedFile('setup.cfg', contents, self):
            lizy = pylease.Pylease(None, None, None)

            ok_(key1 in lizy.config)
            ok_(key2 in lizy.config)
            eq_(lizy.config[key1], val1)
            eq_(lizy.config[key2], val2)

    def test_make_release_must_update_the_version_in_all_provided_files(self):
        version = '0.12'
        file1_name = 'file1'
        file2_name = 'file2'
        file1_contents = textwrap.dedent("""file1
                                            version='{version}'
                                            """)
        file2_contents = textwrap.dedent("""file2
                                            version='{version}'
                                            """)
        config_contents = textwrap.dedent("""
                                          [pylease]
                                          version-files = {}, {}
                                          """.format(file1_name, file2_name))
        setup_py_contents = textwrap.dedent("""
                                            from setuptools import setup
                                            setup(version='{}')
                                            """.format(version))

        with MockedSetupPy(setup_py_contents, self):
            with MockedFile('setup.cfg', config_contents, self):
                with MockedFile(file1_name, file1_contents.format(version=version), self) as file1:
                    with MockedFile(file2_name, file2_contents.format(version=version), self) as file2:
                        main(['make', '--minor'])
                        expected_version = '0.13'
                        eq_(file1.contents(), file1_contents.format(version=expected_version))
                        eq_(file2.contents(), file2_contents.format(version=expected_version))

    def test_pylease_must_load_external_extensions(self):
        extension_package_name = "some_extension"
        extension_class_name = "SomeExtension"
        extension_contents = textwrap.dedent("""
                                             from pylease.ext import Extension
                                             from mock import MagicMock

                                             class {}(Extension):
                                                def load(self):
                                                    pass

                                             """.format(extension_class_name))

        config_contents = textwrap.dedent("""
                                          [pylease]
                                          use-plugins = {}
                                          """.format(extension_package_name))

        with MockedFile(extension_package_name + '.py', extension_contents, self, for_import=True):
            with MockedFile('setup.cfg', config_contents, self):
                lizy = pylease.Pylease(None, None, None)

                ok_('use-plugins' in lizy.config)
                ok_(extension_package_name in sys.modules)

                subclasses = Extension.__subclasses__()
                subclasses_str = []
                for subclass in subclasses:
                    subclasses_str.append(subclass.__name__)

                ok_(extension_class_name, subclasses_str)

    def test_verbose_argument_must_add_debug_handler(self):
        handlers_count = len(pylease.logger.LOGME.handlers)
        pylease.logger.LOGME.debug = MagicMock()

        main(['--verbose', 'status'])

        eq_(len(pylease.logger.LOGME.handlers), handlers_count + 1)

    def test_plugins(self):
        plugin_name = 'some_plugin'
        config_contents = textwrap.dedent("""
                                          [pylease]
                                          use-plugins = {}
                                          """.format(plugin_name))

        with MockedFile(plugin_name + '.py', '', self):
            with MockedFile('setup.cfg', config_contents, self):
                main(['status'])

        ok_(plugin_name in sys.modules)

    def test_init_command_must_initialize_a_new_project(self):
        project_name = 'some_project_name'

        import os
        orig_mkdir = os.mkdir
        os.mkdir = MagicMock()

        orig_listdir = os.listdir
        os.listdir = MagicMock()
        os.listdir.return_value = []

        with patch('pylease.cmd.open', mock_open(), create=True) as open_mock:
            main(['init', project_name])

        os.mkdir = orig_mkdir
        os.listdir = orig_listdir

        handle = open_mock.return_value.__enter__.return_value

        handle.write.assert_any_call(InitCommand.SETUP_PY_CONTENTS.format(name=project_name))
        handle.write.assert_any_call(InitCommand.SETUP_CFG_CONTENTS.format(name=project_name))
        handle.write.assert_any_call(InitCommand.INIT_PY_CONTENTS.format(version=InitCommand.INITIAL_VERSION))



from nose.tools import eq_
from pylease.filemgmt import replace_version, update_files, VersionRollback
from tests import PyleaseTest, MockedFile, MockedSetupPy

__author__ = 'bagrat'


class FileManagementTest(PyleaseTest):
    def test_replace_version_must_replace_a_version_specification_in_the_provided_string(self):
        setup_py_tp = """
                      some line here
                      and another here

                      version = "{version}"

                      and the last one
                      """

        old_version = '0ld.v3rs10n'
        new_version = 'n3w.v3rs10n'

        old_setup_py = setup_py_tp.format(setup_py_tp, version=old_version)
        expected_new_setup_py = setup_py_tp.format(setup_py_tp, version=new_version)

        actual_new_setup_py, _ = replace_version(old_setup_py, old_version, new_version)

        eq_(actual_new_setup_py, expected_new_setup_py, 'replace_version() must update the version')

    def test_replace_version_must_return_the_number_of_occurrences_of_version_specifications(self):
        setup_py_tp = """
                      some line here
                      and another here

                      version = "{version}"

                      and the last one
                      version = "{version}"
                      """

        old_version = '0ld.v3rs10n'
        new_version = 'n3w.v3rs10n'

        old_setup_py = setup_py_tp.format(setup_py_tp, version=old_version)
        expected_new_setup_py = setup_py_tp.format(setup_py_tp, version=new_version)
        expected_count = 2

        actual_new_setup_py, actual_count = replace_version(old_setup_py, old_version, new_version)

        eq_(expected_new_setup_py, actual_new_setup_py, "replace_version() must update version")
        eq_(expected_count, actual_count, "replace_version() must return correct number of occurrences")

    def test_replace_version_must_return_zero_occurrences_if_no_version_specification_found(self):
        setup_py_tp = """
                      some line here
                      and another here

                      and the last one
                      """

        non_existing_text = "some text that does not appear in setup_py"

        _, count = replace_version(setup_py_tp, non_existing_text, 'does not matter')

        eq_(count, 0)

    def test_update_files_must_update_version_in_provided_files(self):
        old_version = 'old_version'
        new_version = 'new_version'

        file_contents = """
                        line one
                        version = {}
                        """

        old_contents = file_contents.format(old_version)
        expected_contents = file_contents.format(new_version)

        filename = 'some_file'

        with MockedFile(filename, old_contents, self) as mocked_file:
            update_files(old_version, new_version, [filename])

            new_contents = mocked_file.contents()

            eq_(expected_contents, new_contents)

        with MockedSetupPy(old_contents, self, for_import=False) as setup_py_mock:
            update_files(old_version, new_version)

            new_contents = setup_py_mock.contents()

            eq_(expected_contents, new_contents)

    def test_update_files_must_update_version_in_setuppy_if_no_files_are_provided(self):
        old_version = 'old_version'
        new_version = 'new_version'

        file_contents = """
                        line one
                        version = {}
                        """

        old_contents = file_contents.format(old_version)
        expected_contents = file_contents.format(new_version)

        with MockedSetupPy(old_contents, self, for_import=False) as setup_py_mock:
            update_files(old_version, new_version)

            new_contents = setup_py_mock.contents()

            eq_(expected_contents, new_contents)

    def test_version_rollback_must_reset_the_old_version(self):
        old_version = 'old_version'
        new_version = 'new_version'

        file_contents = """
                        line one
                        version = {}
                        """

        current_contents = file_contents.format(new_version)
        expected_contents = file_contents.format(old_version)

        with MockedSetupPy(current_contents, self, for_import=False) as setup_py:
            rollback = VersionRollback(old_version, new_version)

            rollback.rollback()

            new_contents = setup_py.contents()

            eq_(expected_contents, new_contents)



import sys
from pylease.cmd.rollback import Rollback, Stage
from pylease.logger import LOGME as logme  # noqa
from pylease.cmd.task import AfterTask
from pylease.ext import Extension


class PypiExtension(Extension):
    # pylint: disable=too-few-public-methods
    def load(self):
        make_command = self._lizy.commands['make']
        parser = make_command.parser

        pypi_group = parser.add_argument_group(title='PyPI arguments')

        pypi_group.add_argument('--pypi',
                                action='store_true',
                                dest='to_pypi',
                                default=False,
                                help='Upload package to PyPI')

        make_command.add_after_task(PypiAfterTask())


class PypiAfterTask(AfterTask):  # pragma: no cover - Unable to test this other than manually TODO: try to test
    # pylint: disable=too-few-public-methods
    def execute(self, lizy, args):
        # pylint: disable=unused-argument
        if args.to_pypi:
            self.enable_rollback(PypiRollback(sys.argv))

            sys.argv = ['setup.py', 'sdist', 'upload']
            self._rollback.enable_stage(PypiRollback.UPLOADING_STAGE)

            __import__('setup')
            self._rollback.enable_stage(PypiRollback.UPLOADED_STAGE)


class PypiRollback(Rollback):  # pragma: no cover
    UPLOADED_STAGE = 'uploaded'
    UPLOADING_STAGE = 'uploading'

    def __init__(self, orig_args):
        super(PypiRollback, self).__init__()

        self._orig_args = orig_args

    @Stage(UPLOADING_STAGE)
    def rollback_args(self):
        sys.argv = self._orig_args

    @Stage(UPLOADED_STAGE)
    def rollback_inform(self):
        # pylint: disable=no-self-use
        logme.warn('Unable to rollback PyPI upload, please delete the uploaded version manually.')



from unittest import TestCase
from nose.tools import ok_, eq_
from pylease.util import get_caller_module, ignore_subclass, IgnoreSubclass


class UtilitiesTest(TestCase):
    def test_subclass_ignore_mark_should_be_true_only_for_marked_class(self):
        class Z(object):
            pass

        @ignore_subclass
        class A1(Z):
            pass

        class B1(A1):
            pass

        class C1(B1):
            pass

        ok_(A1.ignore)
        ok_(not B1.ignore)
        ok_(not C1.ignore)
        ok_(not hasattr(Z, IgnoreSubclass.DEFAULT_MARK_NAME))

        ignore_mark_name = 'ignore_mark'

        @IgnoreSubclass(ignore_mark_name)
        class A2(Z):
            pass

        class B2(A2):
            pass

        class C2(B2):
            pass

        ok_(getattr(A2, ignore_mark_name))
        ok_(not getattr(B2, ignore_mark_name))
        ok_(not getattr(C2, ignore_mark_name))
        ok_(not hasattr(Z, ignore_mark_name))

    def test_get_caller_module_must_return_the_module_from_which_the_method_was_called(self):
        module = get_caller_module()

        eq_(module.__name__, 'tests.test_util')

        def wrapper():
            return get_caller_module(1)

        module = wrapper()

        eq_(module.__name__, 'tests.test_util')



import inspect


def get_caller_module(depth=0):
    """Get the module from where this method was called

    If this method is wrapped into another one, and the client is calling the
    wrapper method, this will return the wrapper method. The solution to this
    situation is to specify the depth when calling this method. For that case
    it should be 1.
    Copied from pyflect 0.1.3, not to have any cross-dependence.

    :param depth: The depth of the call
    :return: The module instance
    """

    frame = inspect.stack()[1 + depth]
    return inspect.getmodule(frame[0])


class SubclassIgnoreMark(object):
    # pylint: disable=too-few-public-methods
    """
    This class is designed to mark particular classes for special purposes. It
    implements the Descriptor protocol, and the value of a property instance of
    this class is True only for the class where the property was originally assigned.
    """
    def __init__(self, cls_name):
        """
        :param cls_name: The class name to be marked.
        """
        super(SubclassIgnoreMark, self).__init__()
        self._cls_name = cls_name

    def __get__(self, instance, owner):
        return owner.__name__ == self._cls_name


def ignore_subclass(klass):
    """
    This decorator may be used in place of `IgnoreSubclass` in order to emit the parenthesis,
    and use the default property name for the Ignore Mark. It does not add any functionality.
    """
    decorator = IgnoreSubclass()

    return decorator(klass)


class IgnoreSubclass(object):
    # pylint: disable=too-few-public-methods
    """
    This decorator adds the Ignore Mark to the decorated class, without a need to specify
    the class name itself. This class requires to define the property name to be used for
    the Ignore Mark. To avoid this verbosity the mark has a default value of `"ignore"`.
    """
    DEFAULT_MARK_NAME = 'ignore'

    def __init__(self, mark_name=DEFAULT_MARK_NAME):
        super(IgnoreSubclass, self).__init__()

        self._mark_name = mark_name

    def __call__(self, klass):
        setattr(klass, self._mark_name, SubclassIgnoreMark(klass.__name__))

        return klass



__author__ = 'bagrat'


def replace_version(content, from_version, to_version):
    """
    Replaces the value of the version specification value in the contents of
    ``contents`` from ``from_version`` to ``to_version``.

    :param content: The string containing version specification
    :param to_version: The new version to be set
    :return: (result_content, number of occurrences)
    """
    frm_str = str(from_version)
    to_str = str(to_version)

    count = content.count(frm_str)

    result_setup_py = content.replace(frm_str, to_str, count)

    return result_setup_py, count


def update_files(from_version, to_version, files=None):
    """
    Update contents of files. See replace_version.

    :param from_version: The old version to be replaced.
    :param to_version: The new version to be set.
    :param files: A list of files, to update versions in.
    """
    counts = {}
    if not files:
        files = ['setup.py']
    for filename in files:
        with open(filename, 'r') as setup_py:
            content = setup_py.read()

            new_content, count = replace_version(content, from_version, to_version)

            counts[filename] = count

        with open(filename, 'w') as setup_py:
            setup_py.write(new_content)

    return counts


class VersionRollback(object):
    # pylint: disable=too-few-public-methods
    """
    A rollback object that simple rolls back the updated version.
    """
    def __init__(self, old_version, new_version, files=None):
        super(VersionRollback, self).__init__()

        self._old = old_version
        self._new = new_version
        self._files = files

    def rollback(self):
        update_files(self._new, self._old, self._files)



from unittest import TestCase
from mock import patch, MagicMock, call
from nose.tools import ok_, eq_
from pylease.ex import PyleaseError
from pylease import Pylease, InfoContainer
from pylease.util import SubclassIgnoreMark
from pylease.cmd import Command, NamedCommand
from pylease.cmd.task import BeforeTask, AfterTask


class CommandTest(TestCase):
    def test_command_global_init_should_ignore_marked_commands(self):
        inited = []

        class C0(Command):
            def __init__(self, dummy):
                super(C0, self).__init__(dummy, 'C0', 'C0 description')

                inited.append(self.__class__)

            def _process_command(self, lizy, args):
                pass

            def _get_name(self):
                pass

        class C1(C0):
            ignore_me = SubclassIgnoreMark('C1')

        class C2(C1):
            pass

        cmd_subparser = MagicMock()
        cmd_subparser.add_parser = MagicMock()
        Command.init_all(Pylease(None, cmd_subparser, None))

        ok_(C0 in inited)
        ok_(C1 not in inited)
        ok_(C2 in inited)

    def test_named_command_class_should_set_command_name_if_not_defined(self):
        description = 'some description'

        class SomeCommand(NamedCommand):
            def __init__(self, lizy):
                super(SomeCommand, self).__init__(lizy, description)

            def _process_command(self, lizy, args):
                return None, None

        subparser_mock = lambda: 0
        subparser_mock.add_parser = MagicMock()

        no_name = SomeCommand(Pylease(None, subparser_mock, None))

        eq_(no_name.name, 'some')
        subparser_mock.add_parser.assert_called_once_with('some', help=description)

    def test_calling_command_instance_should_fire_process_command_method(self):
        class C(Command):
            def __init__(self, lizy):
                super(C, self).__init__(lizy, 'C', 'C description')

            def _process_command(self, lizy, args):
                return None, None

        with patch.object(C, '_process_command') as process_command:
            process_command.return_value = None, None

            subparser_mock = lambda: 0
            subparser_mock.add_parser = MagicMock()
            args_mock = lambda: 0
            args_mock.verbose = False
            info_container = InfoContainer()
            info_container.is_empty = False
            lizy = Pylease(None, subparser_mock, info_container)
            c = C(lizy)
            c(args_mock)

            process_command.assert_called_once_with(lizy, args_mock)

    def test_init_all_should_add_all_commands_to_lizy(self):
        class Base(NamedCommand):
            def __init__(self, lizy, description):
                super(Base, self).__init__(lizy, description)

            ignore_me = SubclassIgnoreMark('Base')

            def _process_command(self, lizy, args):
                pass

        class A(Command):
            NAME = 'A'
            DESC = 'A desc'

            def __init__(self, lizy):
                super(A, self).__init__(lizy, self.NAME, self.DESC)

            def _process_command(self, lizy, args):
                pass

        class BCommand(Base):
            NAME = 'B'
            DESC = 'B desc'

            def __init__(self, lizy):
                super(BCommand, self).__init__(lizy, self.DESC)

        class CCommand(Base):
            NAME = 'C'
            DESC = 'C desc'

            def __init__(self, lizy):
                super(CCommand, self).__init__(lizy, self.DESC)

        subparser_mock = lambda: 0
        subparser_mock.add_parser = MagicMock()

        lizy = Pylease(None, subparser_mock, None)
        Command.init_all(lizy)

        calls = [call(A.NAME, help=A.DESC), call(BCommand.NAME.lower(), help=BCommand.DESC),
                 call(CCommand.NAME.lower(), help=CCommand.DESC)]
        subparser_mock.add_parser.assert_has_calls(calls, any_order=True)

    def test_command_must_execute_before_tasks_before_self_execution(self):
        class SuccessCommand(NamedCommand):
            def __init__(self, lizy):
                super(SuccessCommand, self).__init__(lizy, 'desc')

            def _process_command(self, lizy, args):
                return None, None

        before = MagicMock()
        after = MagicMock()

        lizy = MagicMock()
        args = MagicMock()

        info_container = InfoContainer()
        info_container.is_empty = False

        lizy.info_container = info_container

        sc = SuccessCommand(lizy)

        sc.add_before_task(before)
        sc.add_after_task(after)

        sc(args)

        before.assert_called_once_with(lizy, args)
        after.assert_called_once_with(lizy, args)

    def test_command_must_not_execute_after_tasks_in_case_of_failure(self):
        class FailureCommand(NamedCommand):
            def __init__(self, lizy):
                super(FailureCommand, self).__init__(lizy, 'desc')

            def _process_command(self, lizy, args):
                raise PyleaseError()

        before = MagicMock()
        after = MagicMock()

        lizy = MagicMock()
        args = MagicMock()

        info_container = InfoContainer()
        info_container.is_empty = False
        lizy.info_container = info_container

        sc = FailureCommand(lizy)

        sc.add_before_task(before)
        sc.add_after_task(after)

        sc(args)

        before.assert_called_once_with(lizy, args)
        ok_(not after.called)

    def test_all_actions_must_be_rolled_back_on_failure(self):
        before_rollback = MagicMock()
        command_rollback = MagicMock()
        after_rollback = MagicMock()

        class RollbackTestCommand(NamedCommand):
            def __init__(self, lizy):
                super(RollbackTestCommand, self).__init__(lizy, 'rollback desc', command_rollback)

            def _process_command(self, lizy, args):
                return None

        class B(BeforeTask):
            def __init__(self):
                super(B, self).__init__(before_rollback)
                self.enable_rollback(before_rollback)

            def execute(self, lizy, args):
                pass

        class A(AfterTask):
            def __init__(self):
                super(A, self).__init__(after_rollback)
                self._needs_rollback = True

            def execute(self, lizy, args):
                raise PyleaseError()

        subparser_mock = lambda: 0
        subparser_mock.add_parser = MagicMock()
        args_mock = lambda: 0
        args_mock.verbose = False

        info_container = InfoContainer()
        info_container.is_empty = False
        lizy = Pylease(None, subparser_mock, info_container)
        command = RollbackTestCommand(lizy)

        command.add_before_task(B())
        command.add_after_task(A())

        command(args_mock)

        before_rollback.assert_called_once_with()
        command_rollback.assert_called_once_with()
        after_rollback.assert_called_once_with()



from abc import ABCMeta, abstractmethod
from pylease.logger import LOGME as logme  # noqa


class Extension(object):
    """
    The entry point to implementing Pylease extensions. Pylease loads subclasses of this class and invokes the
    :func:`~pylease.ext.Extension.load` method.

    Attributes:
        _lizy (pylease.Pylease): The :class:`~pylease.Pylease` singleton, that is initialised and passed to all subclass instances.
    """
    # pylint: disable=abstract-class-not-used
    __metaclass__ = ABCMeta

    _IGNORE_ME_VAR_NAME = 'ignore_me'
    ignore_me = False

    def __init__(self, lizy):
        super(Extension, self).__init__()

        logme.debug("Initializing %s", self.__class__.__name__)

        self._lizy = lizy

        self.load()

    @abstractmethod
    def load(self):
        """
        This method is being called by Pylease when all the extensions are being loaded.  All the initialisation code must be implemented
        in the body of this method.
        """
        pass  # pragma: no cover

    @classmethod
    def init_all(cls, lizy):
        extensions = cls.__subclasses__()

        for extension in extensions:
            extension.init_all(lizy)
            if not getattr(extension, cls._IGNORE_ME_VAR_NAME):
                extension(lizy)



from setuptools import setup, find_packages
import pylease

download_url = "https://github.com/n9code/pylease/archive/v{version}.tar.gz".format(version=pylease.__version__)

with open('./deps/test.txt') as req:
    tests_require = req.read().split('\n')

with open('./deps/core.txt') as req:
    install_requires = req.read().split('\n')

with open('./README.rst') as desc:
    long_desc = desc.read()

classifiers = ['License :: OSI Approved :: MIT License',
               'Intended Audience :: Developers',
               'Natural Language :: English',
               'Programming Language :: Python',
               'Programming Language :: Python :: 2',
               'Programming Language :: Python :: 3',
               'Operating System :: POSIX',
               'Operating System :: POSIX :: Linux',
               'Operating System :: MacOS',
               'Topic :: Software Development :: Libraries :: Python Modules']

config = {
    'name': 'pylease',
    'description': 'Easy package versioning and release management',
    'long_description': long_desc,
    'author': 'Bagrat Aznauryan',
    'url': 'git@github.com:n9code/pylease.git',
    'download_url': download_url,
    'author_email': 'bagrat@aznauryan.org',
    'version': pylease.__version__,
    'install_requires': install_requires,
    'tests_require': tests_require,
    'classifiers': classifiers,
    'packages': find_packages(),
    'license': 'MIT',
    'keywords': 'release version versioning',
    'entry_points': {
        'console_scripts': [
            'pylease = pylease.main:main'
        ]
    }
}

setup(**config)



from mock import MagicMock
from nose.tools import eq_, ok_
from pylease.cmd.rollback import Rollback, Stage
from tests import PyleaseTest

stage1 = 'stage1'
stage1_priority = 0
stage2 = 'stage2'
stage2_priority = 2
stage3 = 'stage3'
stage3_priority = 1
stage1_mock = MagicMock()
stage2_mock = MagicMock()
stage3_mock = MagicMock()
stage_execution_queue = []


class RollbackTest(PyleaseTest):
    def setUp(self):  # noqa
        super(RollbackTest, self).setUp()
        global stage_execution_queue

        stage_execution_queue = []

    class SomeRollback(Rollback):
        @Stage(stage1, stage1_priority)
        def stage1_rollback(self):
            stage1_mock()
            stage_execution_queue.append(stage1_mock)

        @Stage(stage2, stage2_priority)
        def stage2_rollback(self):
            stage2_mock()
            stage_execution_queue.append(stage2_mock)

        @Stage(stage3, stage3_priority)
        def stage3_rollback(self):
            stage3_mock()
            stage_execution_queue.append(stage3_mock)

    def test_rollback_mechanism_should_generate_appropriate_dictionary(self):
        rollback = self.SomeRollback()

        expected_stages = {stage1: {rollback.stage1_rollback: (stage1_priority, False)},
                           stage2: {rollback.stage2_rollback: (stage2_priority, False)},
                           stage3: {rollback.stage3_rollback: (stage3_priority, False)}}

        eq_(expected_stages, rollback._stages)

    def test_it_must_be_possible_to_cherry_pick_enable_rollback_stages(self):
        rollback = self.SomeRollback()

        rollback.enable_stage(stage2)

        ok_(rollback._stages[stage2][rollback.stage2_rollback][1])
        ok_(not rollback._stages[stage1][rollback.stage1_rollback][1])

    def test_rollback_mechanism_should_execute_enabled_stages_in_correct_order(self):
        rollback = self.SomeRollback()

        rollback.enable_stage(stage2)
        rollback.enable_stage(stage3)
        rollback.rollback()

        ok_(stage1_mock not in stage_execution_queue)
        stage2_mock.assert_called_once_with()
        stage3_mock.assert_called_once_with()

        stage2_index = stage_execution_queue.index(stage2_mock)
        stage3_index = stage_execution_queue.index(stage3_mock)

        ok_(stage2_index < stage3_index)



from distutils.version import StrictVersion
import re
import string

__author__ = 'bagrat'


class DevedVersion(StrictVersion):
    """
    An extension to StrictVersion, which includes a dev component.
    """
    _strict_re_str = StrictVersion.version_re.pattern[1:-1]
    _deved_re_str = r'^({strict_re}) (\.dev(\d+))?$'\
        .format(strict_re=_strict_re_str)
    _deved_re = re.compile(_deved_re_str, re.VERBOSE)

    def __init__(self, vstring=None):
        match = self._deved_re.match(vstring)

        if not match:
            raise ValueError("Invalid version number '%s'" % vstring)
        else:
            dev_num_index = len(match.groups())
            if match.group(dev_num_index):
                self.dev = string.atoi(match.group(dev_num_index))
            else:
                self.dev = 0
            vstring = match.group(1)

        self.version = None

        StrictVersion.__init__(self, vstring)

    def _increase_version(self, count=1, index=0):
        """
        Increase the index'th level of the version by count
        """
        old_version = self.version

        old_comp = old_version[index]
        new_comp = old_comp + count

        new_version = []

        for i, _ in enumerate(old_version):
            if i == index:
                new_version.append(new_comp)
            elif i > index:
                new_version.append(0)
            else:
                new_version.append(old_version[i])

        self.dev = 0
        self.version = tuple(new_version)

    def increase_major(self, count=1):
        """
        Increase the major version by count
        """
        self._increase_version(count, 0)

    def increase_minor(self, count=1):
        """
        Increase the minor version by count
        """
        self._increase_version(count, 1)

    def increase_patch(self, count=1):
        """
        Increase the patch version by count
        """
        self._increase_version(count, 2)

    def increase_dev(self, level=1):
        """
        Increase the dev version by count
        """
        self.dev += level

    def __str__(self):
        prefix = StrictVersion.__str__(self)
        suffix = '.dev{ver}'.format(ver=self.dev) if self.dev > 0 else ''

        return '{prefix}{suffix}'.format(prefix=prefix, suffix=suffix)





class Rollback(object):
    """
    This class provides a facility to define a staged rollback process. The scenario of using this class is the following:

        #. Inherit :class:`~pylease.cmd.rollback.Rollback` class
        #. Define rollback stages as instance methods
        #. Decorate each rollback method with :class:`~pylease.cmd.rollback.Stage` decorator, by specifying stage name and priority
        #. Enable each stage separately calling :func:`~pylease.cmd.rollback.Rollback.enable_stage` method
    """
    def __init__(self):
        super(Rollback, self).__init__()

        self._stages = {}

        for rollback in dir(self):
            rollback_obj = getattr(self, rollback)
            if hasattr(rollback_obj, Stage.STAGE_ATTR_NAME):
                stage = getattr(rollback_obj, Stage.STAGE_ATTR_NAME)
                priority = getattr(rollback_obj, Stage.PRIORITY_ATTR_NAME)

                if not self._stages.get(stage):
                    self._stages[stage] = {}

                self._stages[stage][rollback_obj] = (priority, False)  # (priority, enabled)

    def rollback(self):
        """
        Execute all rollback stages ordered by priority.
        """
        rollbacks = []
        for stage in self._stages:
            for rollback in self._stages[stage]:
                priority, enabled = self._stages[stage][rollback]

                if enabled:
                    rollbacks.append((stage, priority, rollback))

        rollbacks = sorted(rollbacks, key=lambda obj: obj[1], reverse=True)

        for rollback in rollbacks:
            rollback[2]()

    def enable_stage(self, stage):
        """
        Enable particular stage by name.

        Arguments:
            stage (str): Stage name to enable.
        """
        rollbacks = self._stages[stage]

        for rollback in rollbacks:
            priority, _ = self._stages[stage][rollback]
            self._stages[stage][rollback] = (priority, True)

    def __call__(self):
        self.rollback()  # pragma: no cover - For simple use and convenience


class Stage(object):
    # pylint: disable=too-few-public-methods
    """
    Decorator used in custom :class:`~pylease.cmd.rollback.Rollback` classes for associating each method with a stage, and setting priority.

    Arguments:
        stage (str): The name of the stage.
        priority (int): The order priority of the stage to be rolled back. Defaults to ``0``.

    Example:
        Here is an example of how to use the :class:`~pylease.cmd.rollback.Stage` decorator in combination with the
        :class:`~pylease.cmd.rollback.Rollback` base class::

            class ExampleRollback(Rollback):
                @Stage('some_stage', 1)
                def some_stage_with_priority_1(self):
                    pass  # your some_stage rollback goes here

    """
    STAGE_ATTR_NAME = '_stage'
    PRIORITY_ATTR_NAME = '_priority'

    def __init__(self, stage, priority=0):
        super(Stage, self).__init__()

        self._stage = stage
        self._priority = priority

    def __call__(self, func):
        setattr(func, self.STAGE_ATTR_NAME, self._stage)
        setattr(func, self.PRIORITY_ATTR_NAME, self._priority)

        return func
